#!/usr/bin/perl -w
#
# description: Monitor NFS for stale and hung filesystems

my $confEmail = "root";
#my $confEmail  = "olchansk";
my $confState  = "/var/run/monitor_nfs.txt";
my $confStatfs = "/usr/sbin/monitor_nfs_statfs";
#my $confStatfs = "/home/olchansk/sysadm/rpm/monitor_nfs/monitor_nfs_statfs";

die "\"$confStatfs\" is not executable" unless -x $confStatfs;

#my $verbose = grep(/-v/,@ARGV);
#my $dryrun  = grep(/-n/,@ARGV);

my $nosleep  = grep(/--nosleep/,@ARGV);

if (!$nosleep) {
  my $r = rand(15*60);
  #print "sleep $r sec\n";
  sleep $r;
}

read_state();
monitor();

exit 0;

sub read_state
  {
    return unless -r $confState;
    open(TMPIN,$confState);
    while (my $in = <TMPIN>)
      {
        my ($d,$mnt,$p) = split(/\s+/,$in);
        $since{$mnt} = $d;
      }
    close TMPIN;
  }

sub monitor
  {
    my @mnt;
    my %dev;

    my $now = time();
    my $host = `hostname`; chop $host;

    open (TMPIN,"/proc/mounts");
    while (my $in = <TMPIN>)
      {
        my ($dev,$mnt,$fs) = split(/\s+/,$in);
        next if $fs eq "nfsd";
        next unless $fs =~ /nfs/;
        #print "[$dev] [$mnt] [$fs]\n";
        push @mnt,$mnt;
        $dev{$mnt} = $dev;
      }
    close TMPIN;

    my %problem;

    foreach my $mnt (sort @mnt)
      {
        $problem{$mnt} = check_nfs($dev{$mnt},$mnt);
      }

    my $report = "";
    my $hostlist;
    my $count = 0;
    foreach my $mnt (sort keys %problem)
      {
        my $p = $problem{$mnt};

        my $d = $since{$mnt};
        if (! defined $d)
          {
            $d = $now;
            $since{$mnt} = $d;
          }

        my $dd = localtime($d);
        next unless $p;
        $count ++;
        $report .= "$mnt: $p, fs bad since $dd\n";

        my ($host,$export) = split(/\:/,$dev{$mnt});
        $hostlist{$host} = $host;
      }

    if ($count > 0)
      {
        my $hostlist = join(" ", keys %hostlist);
        send_email($confEmail,"$host: $count nfs problems: $hostlist",$report);
      }

    if (-w $confState)
      {
        open (TMPOUT,">$confState");
        foreach my $mnt (sort keys %problem)
          {
            my $p = $problem{$mnt};
            next unless $p;
            my $d = $since{$mnt};
            print TMPOUT "$d $mnt $p\n";
          }
      }
    close TMPOUT;
  }

sub check_nfs
  {
    my $dev = shift @_;
    my $mnt = shift @_;

    my ($host,$export) = split(/\:/,$dev);

    #return "\"$host\" does not ping" unless ping($host);

    my $chk = check_hung_statfs($confStatfs,$mnt);
    return $chk if $chk;

    my $df = `$confStatfs \'$mnt\' 2>&1`;
    return "statfs() timeout, \"$dev\" is hung" if $df =~ /Alarm/;
    return  if $df =~ /Value too large for defined data type/; # ignore bizarre 2.6 vs 2.4 NFS interaction
    return "Strange statfs() results: $df" unless $df =~ /f_type/;

    #print "$dev: $df\n";

    return;
  }

sub check_hung_statfs
  {
    my $exe = shift @_;
    my $mnt = shift @_;

    opendir (DIR,"/proc");
    while (my $dir = readdir(DIR))
      {
        my $link = readlink "/proc/$dir/exe";
        next if ! defined $link;
        next unless ($link =~ /$exe/);
        my $cmdline = `/bin/cat /proc/$dir/cmdline 2>&1`;
        $cmdline =~ s/\0/XXX/g;
        next unless ($cmdline =~ m!$mnt!m);
        #print "mnt $mnt, dir $dir, link $link, cmd: $cmdline\n";
        return "Stuck \"$exe $mnt\" with pid $dir";
      }
    closedir DIR;

    return;
  }

sub ping
  {
    my $host = shift @_;
    system ("/bin/ping -c 1 -w 2 $host &> /dev/null") ;
    my $stat = $?;
    return 0 if ($stat);
    return 1;
  }

sub send_email
  {
    my $dest    = shift @_;
    my $subject = shift @_;
    my $text    = shift @_;

    open(MAIL,"|mail -s \"$subject\" $dest");
    print MAIL "$text\n";
    close MAIL;
  }

#end
